#lang racket/gui
(require data/heap)
; priority queue and helper functions

(define (comp n1 n2)
  (let ([d1 (cdr n1)]
        [d2 (cdr n2)])
    (<= d1 d2)))

(define queue (make-heap comp))

(define (enqueue n) (heap-add! queue n))

(define (dequeue)
  (let ([n (heap-min queue)])
    (heap-remove-min! queue)
    n))
            
; Puzzle Structure ------------------------------------------

(define SIZE 4)
(define empty (sqr SIZE))
  
(define (ref puzzle r c)
  (let ([i (+ c (* r SIZE))])
    (vector-ref puzzle i)))

(define tile-homes
  (let ([hash (make-hash)])
    (for ([n (in-range (sqr SIZE))])
      (hash-set! hash (add1 n) (cons (quotient n SIZE) (remainder n SIZE))))
    hash))
  
(define (manhattan puzzle) 
  (let ([dist 0])
    (for* ([r (in-range SIZE)]
           [c (in-range SIZE)])
      (let ([t (ref puzzle r c)])
        (unless (= empty t)
          (let* ([th (hash-ref tile-homes t)]
                 [hr (car th)]
                 [hc (cdr th)])
            (set! dist (+ dist
                          (abs (- hr r))
                          (abs (- hc c))))))))
    dist))

(define (zone1a puzzle) 
  (let ([dist 0])
    (for* ([r (in-range SIZE)]
           [c (in-range SIZE)])
      (let ([t (ref puzzle r c)])
        (unless (= t empty)
          (let* ([th (hash-ref tile-homes t)]
                 [hr (car th)]
                 [hc (cdr th)]
                 [d (+ (abs (- hr r)) (abs (- hc c)))])
            (when (or (and (<= t 4) (> r 1)))
              (set! dist (+ dist d)))))))
    dist))

#;(define (zone1a puzzle) 
    (let ([dist 0])
      (for* ([r (in-range SIZE)]
             [c (in-range SIZE)])
        (let ([t (ref puzzle r c)])
          (unless (= t empty)
            (let* ([th (hash-ref tile-homes t)]
                   [hr (car th)]
                   [hc (cdr th)]
                   [d (+ (abs (- hr r)) (abs (- hc c)))])
              (when (or (and (<= t 4) (> r 0 )) (and (> t 4) (= 0 r)))
                (set! dist (+ dist d)))))))
      dist))

(define (zone1b puzzle) 
  (let ([dist 0])
    (for* ([r (in-range SIZE)]
           [c (in-range SIZE)])
      (let* ([t (ref puzzle r c)])
        (when (<= t 4)
          (let* ([th (hash-ref tile-homes t)]
                 [hr (car th)]
                 [hc (cdr th)]
                 [d (+ (abs (- hr r)) (abs (- hc c)))])
            (set! dist (+ dist d))))))
    dist))

(define zone2-tiles (set 5 9 13))

(define (zone2 puzzle) 
  (let ([dist 0])
    (for* ([r (in-range 1 SIZE)]
           [c (in-range SIZE)])
      (let* ([t (ref puzzle r c)])
        (when (set-member? zone2-tiles t)
          (let* ([th (hash-ref tile-homes t)]
                 [hr (car th)]
                 [hc (cdr th)]
                 [d (+ (abs (- hr r)) (abs (- hc c)))])
            (set! dist (+ dist d))))))
    dist))

(define (move-offset i)
  (case i
    [(0) (values  0 -1)]
    [(1) (values  0  1)] 
    [(2) (values -1  0)] 
    [(3) (values  1  0)]))

(define (empty-loc puzzle)
  (values
   (vector-ref puzzle empty)
   (vector-ref puzzle (add1 empty))))

(define (make-move puzzle i)
  (let*-values ([(ro co) (move-offset i)]
                [(re ce) (empty-loc puzzle)]
                [(rt ct) (values (+ re ro) (+ ce co))]
                [(t) (ref puzzle rt ct)])
    (for/vector ([i (in-range (+ 2 (sqr SIZE)))])
      (cond [(< i empty)
             (let-values ([(r c) (quotient/remainder i SIZE)])
               (cond [(and (= r re) (= c ce)) t]
                     [(and (= r rt) (= c ct)) empty]
                     [else (vector-ref puzzle i)]))]
            [(= i empty) rt]
            [else ct]))))

(define min-r 0)
(define min-c 0)

(define (next-states puzzle)
  (let-values ([(re ce) (empty-loc puzzle)])
    (define (legal i)
      (let*-values ([(ro co) (move-offset i)]
                    [(rt ct) (values (+ re ro) (+ ce co))])
        (and (>= rt min-r) (>= ct min-c) (< rt SIZE) (< ct SIZE))))
    (for/list ([i (in-range 4)] #:when (legal i))
      (make-move puzzle i))))
       
(define (print puzzle)
  (for* ([r (in-range SIZE)]
         [c (in-range SIZE)])
    (when (= 0 c) (printf "\n"))
    (let ([t (ref puzzle r c)])
      (if (= t empty)
          (printf "   ")
          (printf " ~a" (~a t #:min-width 2 #:align 'right) ))))
  (printf "\n"))

; ---------------------------------------------------
(define closed (make-hash))

(define (solution-list n l)
  (if (equal? n null)
      l
      (let* ([parent (first (hash-ref closed n))])
        (solution-list parent (cons n l)))))

(define (print-solution l)
  (for ([p l]) (print p)))

(define (a-star puzzle fscore)
  (let ([solution null]
        [goal null])
    (hash-clear! closed)
    (set! queue (make-heap comp))  ; open
    (enqueue (cons (list puzzle 0 null) (fscore puzzle)))
    (let loop ()
      (when (equal? solution null)
        (let* ([node-info (dequeue)])
          (match node-info
            [(cons (list node node-depth node-parent) score)
             (if (= 0 (fscore node))
                 (begin
                   (set! goal node)
                   (set! solution (solution-list node-parent (list node))))
                 (begin
                   (hash-set! closed node (list node-parent score)) 
                   (for ([child (next-states node)])
                     (let* ([depth (add1 node-depth)]
                            [next-score (+ depth (fscore child))]
                            [next (cons (list child depth node) next-score)])
                       (if (hash-has-key? closed child)
                           (let* ([prior-score (second (hash-ref closed child))])
                             (when (< next-score prior-score)
                               (hash-remove! closed child)
                               (enqueue next)))
                           (enqueue next))))))])
          (if (> (heap-count queue) 0)
              (loop)
              (when (equal? solution null) (printf "No solution found\n"))))))
    (values goal solution)))

(define (solve puzzle)  
  (set! min-r 0)
  (set! min-c 0)
  (let*-values ([(goal sol-z1a) (a-star puzzle zone1a)])
    (let*-values  ([(goal sol2) (a-star goal zone1b)]
                   [(sol-z1b) (cdr sol2)])
      (set! min-r 1)
      (let*-values  ([(goal sol3) (a-star goal zone2)]
                     [(sol-z2) (cdr sol3)])
        (set! min-c 1)
        (let*-values  ([(goal sol4) (a-star goal manhattan)]
                       [(sol-man) (cdr sol4)])
          (append sol-z1a sol-z1b sol-z2 sol-man))))))

; GUI stuff --------------------------------------------------------

(define TILE-SIZE 60)
(define INNER-SIZE (* SIZE TILE-SIZE))
(define THICKNESS 10) ; puzzle frame thickness
(define PUZZLE-SIZE (+ INNER-SIZE (* 2 THICKNESS)))
(define MARGIN 10)
(define TIMER-DELAY 1000)
(define SWAPS 16) ; must be even number

(define standard #(1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16 3 3))  
(define solution (list standard))
(define current-state standard)
(define valid-puzzle #t)
(define running #f)

(define main-frame
  (new frame%
       [label "15-Puzzle"]
       [width (+ PUZZLE-SIZE (* 2 (+ MARGIN THICKNESS)))]))

(define main-panel (new vertical-panel%
                        [parent main-frame]))

(define canvas
  (new canvas%
       [parent main-panel]
       [min-height (+ PUZZLE-SIZE (* 2 MARGIN))]
       [paint-callback (lambda (canvas dc) (update-canvas dc))]))

(define control-panel1
  (new horizontal-panel%
       [parent main-panel]))

(define control-panel2
  (new horizontal-panel%
       [parent main-panel]))

(define text-puzzle
  (new text-field%
       [parent control-panel1]
       [label "Puzzle: "]
       (init-value "")))

(define msg
  (new message%
       [parent control-panel1]
       [min-width 40]
       [label "Ready"]))

(define btn-load
  (new button%
       [parent control-panel2]
       [label "Load"]
       [callback (lambda (button event)
                   (load-puzzle (send text-puzzle get-value)))]))

(define btn-mix
  (new button%
       [parent control-panel2]
       [label "Random"]
       [callback (lambda (button event) (random-puzzle))]))

(define btn-step
  (new button%
       [parent control-panel2]
       [label "Step"]
       [callback (lambda (button event) (do-step))]))

(define btn-solve
  (new button%
       [parent control-panel2]
       [label "Solve"]
       [callback (lambda (button event)
                   (if running
                       (begin
                         (send timer stop)
                         (set! running #f)
                         (send btn-solve set-label "Solve")
                         (send msg set-label "Ready"))
                       (begin
                         (send msg set-label "--!--")
                         (send btn-solve set-label "Pause")
                         (set! running #t)
                         (send timer start TIMER-DELAY))))]))

(define (update-canvas dc)
  (if (or (not valid-puzzle) (equal? solution null))
      (let ([font (make-object font% 10 'default 'normal 'bold)])
        (send dc set-font font)
        (send dc set-text-foreground "blue")
        (if (not valid-puzzle)
            (send dc draw-text "Invalid input puzzle." 10 10 )
            (send dc draw-text "No solution found." 10 10 )))
      (draw-puzzle dc)))

(define (draw-puzzle dc)
  (let ([start-x MARGIN]
        [start-y MARGIN])
    (send dc set-smoothing 'smoothed) 
    (send dc set-brush (new brush% [color "dark gray"]))
    (send dc draw-rectangle start-x start-y PUZZLE-SIZE PUZZLE-SIZE)

    (send dc set-brush (new brush% [color "black"]))
    (send dc draw-rectangle (+ start-x THICKNESS) (+ start-y THICKNESS)
          INNER-SIZE INNER-SIZE)
    (draw-tiles dc current-state)
    ))

(define (draw-centered-text dc text x y)
  (let-values ([(w h d s) (send dc get-text-extent text)])
    (let ([x (+ x (/ (- TILE-SIZE w) 2))]
          [y (+ y (/ (- TILE-SIZE h d) 2))])
      (send dc draw-text text x y ))))

(define (draw-tiles dc puzzle)
  (let ([c1 (new brush% [color "green"])]
        [c2 (new brush% [color "yellow"])]
        [pattern #( 0 0 1 0 1 1 0 1 0 0 1 0 1 1 0 1 0 )]
        [font (make-object font% 20 'default 'normal 'bold)])
    (send dc set-font font)
    (send dc set-text-foreground "blue")
    (for ([i (in-range empty)])
      (let*-values ([(r c) (quotient/remainder i SIZE)]
                    [(x y) (values
                            (+ THICKNESS MARGIN (* c TILE-SIZE))
                            (+ THICKNESS MARGIN (* r TILE-SIZE)))]
                    [(n) (vector-ref puzzle i)]
                    [(t) (number->string n)]
                    [(color) (if (= (vector-ref pattern n) 0) c1 c2)])
        (unless (= n empty)
          (send dc set-brush color)
          (send dc draw-rectangle x y TILE-SIZE TILE-SIZE)
          (draw-centered-text dc t x y ))))))

(define (load-puzzle str)
  (set! current-state null)
  (send btn-solve set-label "Solve")
  (set valid-puzzle #t)
  (let ([l (string-split str)]
        [valid #f])
    (when (= empty (length l)) ; correct length
      (let* ([vals (map string->number l)]
             [all-numeric (for/and ([n vals]) (integer? n))])
        (when all-numeric
          (let* ([valid-range (for/and ([n vals]) (and (> n 0) (<= n empty)))])
            (when (and valid-range (member empty vals))
              (let*-values ([(i) (index-of vals empty)]
                            [(re co) (quotient/remainder i SIZE)])
                (set! current-state (list->vector (append vals (list re co))))
                (send canvas refresh-now)
                (send msg set-label "--?--")
                (set! solution (solve current-state))
                (unless (equal? solution null)
                  (send text-puzzle set-value "")
                  (set! valid #t)
                  (set! solution (cdr solution)))
                (send msg set-label "Ready")))))))
    (set! valid-puzzle valid)
    (send canvas refresh-now)))

(define (random-puzzle)
  (let ([puzzle (vector-copy standard)])
    (send msg set-label "--?--")
    (send btn-solve set-label "Solve")
    (send text-puzzle set-value "")
    (let loop ([c SWAPS])
      (if (= c 0)
          (begin
            (set! current-state puzzle)
            (send canvas refresh-now)
            (set! solution (solve puzzle))
            (unless (equal? solution null)
              (set! solution (cdr solution)))
            (send msg set-label "Ready")
            (set! valid-puzzle #t)
            (send canvas refresh-now))
          (let ([i1 (random (sub1 empty))]
                [i2 (random (sub1 empty))])
            (if (= i1 i2)
                (loop c)
                (begin
                  (let-values ([(t1 t2)
                                (values
                                 (vector-ref puzzle i1)
                                 (vector-ref puzzle i2))])
                    (vector-set! puzzle i1 t2)
                    (vector-set! puzzle i2 t1)
                    (loop (sub1 c))))))))))
              
(define (do-step)
  (unless (equal? solution #t)
    (set! current-state (car solution))
    (set! solution (cdr solution))
    (when (equal? solution null)
      (send msg set-label "Done!")
      (send timer stop)
      (send btn-solve set-label "Solve")
      (set! solution #t)
      (set! running #f))
    (send canvas refresh-now)))

(define timer
  (new timer% [notify-callback do-step]))

(send main-frame show #t)

;(a-star #(2 3 1 4 5 6 7 8 9 10 11 12 13 14 15 16 3 3))
;(a-star #(2 10 8 3 1 6 16 4 5 9 7 11 13 14 15 12 1 2))
;(a-star #(5 1 2 4 14 9 3 7 13 10 12 6 15 11 8 16 3 3))