#lang racket/gui

; constants
(define MAX-DISKS 9)
(define UNIT 15)
(define PEG-X (+ (* 3 UNIT) (* MAX-DISKS (/ UNIT 2))))
(define PEG-Y (* 2 UNIT))
(define START-Y (+ PEG-Y (* UNIT MAX-DISKS)))
(define PEG-DIST (* UNIT (add1 MAX-DISKS)))
(define RADIUS (/ UNIT 2))
(define ANIMATION-INTERVAL 1) ; ms
(define MOVE-DIST 2)

; state variables
(define num-disks 8)
(define delta-x 0) 
(define delta-y 0)
(define target-x 0) 
(define target-y 0)
(define source-peg 0)
(define dest-peg 0)
(define current-disk 0)
(define current-x 0)
(define current-y 0)
(define peg-disks (make-vector 3 0))
(define move-list '())
(define total-moves 0)
(define move-num 0)
(define in-motion #f)
(define mode 'stoppd)

; GUI ------------------------
(define main-frame
  (new frame%
       [label "Tower of Hanoi"]
       [width (+ (* 7 UNIT) (* 3 PEG-DIST))]))

(define main-panel (new vertical-panel%
                        [parent main-frame]))

(define canvas
  (new canvas%
       [parent main-panel]
       [min-height (+ START-Y UNIT)]
       [paint-callback (lambda (canvas dc) (update-canvas dc))]))

(define control-panel1
  (new horizontal-panel%
       [parent main-panel]))

(define control-panel2
  (new horizontal-panel%
       [parent main-panel]))

(define btn-reset
  (new button%
       [parent control-panel1]
       [label "Reset"]
       [callback (lambda (button event)
                   (when (not in-motion) (reset)))]))

(define text-disks
  (new text-field%
       [parent control-panel1]
       [label "Disks: "]
       [stretchable-width #f]
       (init-value "8")))

(define slider-disks
  (new slider%
       [parent control-panel1]
       [label #f]
       [stretchable-width #f]
       [min-width 100]
       [style (list' horizontal 'plain)]
       [min-value 1]
       [max-value MAX-DISKS]
       [init-value 8]
       [callback (lambda (slider event)
                   (send text-disks
                         set-value
                         (number->string (send slider-disks get-value))))]))

(define gauge
  (new gauge%
       [parent control-panel1]
       [label "Progress: "]
       [stretchable-width #f]
       [min-width 100]
       (range 100)))

(void (new button%
     [parent control-panel2]
     [label "Step"]
     [callback (lambda (button event)
                 (when (not in-motion)
                   (when (equal? move-list '()) (reset))
                   (set! in-motion #t)
                   (set! mode 'step)
                   (send msg set-label "Running...")
                   (init-next-step)
                   (send timer start ANIMATION-INTERVAL)))]))

(new button%
     [parent control-panel2]
     [label "Solve"]
     [callback (lambda (button event)
                 (when (not in-motion)
                   (let ([old num-disks]
                         [new (validate-disks)])
                     (when (or (equal? move-list '()) (not (= old new)))
                       (set! num-disks new)
                       (reset))
                     (set! in-motion #t)
                     (set! mode 'solve)
                     (send msg set-label "Running...")
                     (init-next-step)
                     (send timer start ANIMATION-INTERVAL))))])

(new button%
     [parent control-panel2]
     [label "Pause"]
     [callback (lambda (button event)
                 (set! mode 'step))])

(define msg
  (new message%
       [parent control-panel2]
       [auto-resize #t]
       [label "Ready"]))
 
; Solve --------------------------------

(define (hanoi n f t)
  (if (= 1 n) (list (list f t))
      (let* ([u (- 3 (+ f t))] ; determine unused peg
             [m1 (hanoi (sub1 n) f u)] ; move n-1 disks from f to u
             [m2 (list f t)] ; move single disk from f to t
             [m3 (hanoi (sub1 n) u t)]); move disks from u to t
        (append m1 (cons m2 m3)))))

; Canvas update ------------------------

(define (disk-y pos)
  (- START-Y (* pos UNIT)))

(define (disk-x n peg)
  (let ([w (* (add1 n) UNIT)])
    (- (+ PEG-X (* peg PEG-DIST)) (/ w 2) (/ UNIT -2))))

(define (make-gradient start stop c1 c2)
  (new linear-gradient%
       [x0 start] [y0 0]
       [x1 stop] [y1 0]
       [stops
        (list (list 0   (make-object color% c1))
              (list 0.5 (make-object color% c2))
              (list 1   (make-object color% c1)))]))
       
(define (draw-disk dc n peg pos)
  (let* ([w (* (add1 n) UNIT)]
         [x (disk-x n peg)]
         [y (disk-y pos)])
    (send dc set-brush
          (new brush%
               [gradient (make-gradient x (+ x w) "Green" "GreenYellow")]))
    (send dc draw-rounded-rectangle x y w UNIT RADIUS)))

(define (draw-peg-disks dc peg disks)
  (define (loop disks pos)
    (when (> pos 0)
      (let ([n (first disks)]
            [r (rest disks)])
        (draw-disk dc n peg pos)
        (loop r (sub1 pos)))))
  (loop disks (length disks)))
      
(define (draw-disks dc)
  (for ([peg (in-range 3)])
    (draw-peg-disks dc peg (vector-ref peg-disks peg))))

(define (draw-base dc)
  (for ([i (in-range 3)])
    (let ([x (+ PEG-X (* i PEG-DIST))])
      (send dc set-brush
            (new brush%
                 [gradient (make-gradient x (+ x UNIT) "Chocolate" "DarkOrange")]))
      (send dc draw-rounded-rectangle x PEG-Y UNIT (+ UNIT (- START-Y PEG-Y)) RADIUS)))
  (send dc set-brush (new brush% [color "Chocolate"]))
  (send dc draw-rectangle UNIT START-Y (+ (* 4 UNIT) (* 3 PEG-DIST)) UNIT))

(define (update-canvas dc)
  (draw-base dc)
  (draw-disks dc)
  (when (current-disk . > . 0)
    (let* ([w (* (add1 current-disk) UNIT)]
           [x current-x]
           [y current-y])
      (send dc set-brush
            (new brush%
                 [gradient (make-gradient x (+ x w) "Green" "GreenYellow")]))
      (send dc draw-rounded-rectangle x y w UNIT RADIUS))))

(define (init-next-step)
  (let ([move (first move-list)])
    (set! source-peg (first move))
    (set! dest-peg (second move))
    (set! delta-x 0)
    (set! delta-y (- MOVE-DIST))
    (set! target-y (/ UNIT 2))
    (set! move-list (rest move-list))
    (let* ([source-disks (vector-ref peg-disks source-peg)]
           [pos (length source-disks)])
      (set! current-disk (first source-disks))
      (set! current-x (disk-x current-disk source-peg))
      (set! current-y (disk-y pos))
      (vector-set! peg-disks source-peg
                   (rest source-disks)))))

(define (validate-disks)
  (let* ([disks-str (send text-disks get-value)]
         [n (string->number disks-str)])
    (if (and (integer? n) (< 0 n (add1 MAX-DISKS)))
        (begin
          (send slider-disks set-value n)
          (send msg set-label "  Ready")
          n)
        (begin
          (send text-disks set-value (number->string num-disks))
          (send msg set-label "  Disks out of range.")
          num-disks))))

(define (update-progress)
  (send gauge set-value (inexact->exact (floor (* 100 (/ move-num total-moves))))))

(define (move-disk)
  (cond [((abs delta-y) . > . 0)
         (begin
           (set! current-y (+ current-y delta-y))
           (when ((abs (- target-y current-y)) . < . MOVE-DIST)
             (set! current-y target-y)
             (if (delta-y . < . 0) 
                 (begin ; moving up
                   (set! target-x (disk-x current-disk dest-peg))
                   (set! delta-x (sgn (- dest-peg source-peg)))
                   (set! delta-y 0))
                 (begin ; moving down
                   (set! move-num (add1 move-num))
                   (vector-set! peg-disks dest-peg
                                (cons current-disk (vector-ref peg-disks dest-peg)))
                   (if (equal? mode 'step)
                       (begin
                         (send timer stop)
                         (set! current-disk 0)
                         (set! in-motion #f)
                         (set! mode 'stopped)
                         (send msg set-label "Ready")
                         (set! delta-y 0))
                       (if (> (length move-list) 0)
                           (init-next-step)
                           (begin
                             (send timer stop)
                             (send msg set-label "Done!")
                             (set mode 'stopped)
                             (set! in-motion #f)
                             (set! delta-y 0))))))))]
        
        [((abs delta-x) . > . 0)
         (begin
           (set! current-x (+ current-x delta-x))
           (when ((abs (- target-x current-x)) . < . MOVE-DIST)
             (set! current-x target-x)
             (set! target-y (* PEG-DIST (- dest-peg source-peg)))
             (set! delta-y MOVE-DIST)
             (let ([tdisks (length (vector-ref peg-disks dest-peg))])
               (set! target-y (disk-y (add1 tdisks))))
             (set! delta-x 0)))]
        
        [else (send timer stop)])
  (update-progress)
  (send canvas refresh-now))
          
(define (reset)
  (set! num-disks (validate-disks))
  (set! delta-x 0)
  (set! delta-y 0)
  (set! current-disk 0)
  (set! move-list (hanoi num-disks 0 2))
  (set! total-moves (length move-list))
  (set! move-num 0)
  (vector-set! peg-disks 0 (range 1 (+ 1 num-disks)))
  (vector-set! peg-disks 1 '())
  (vector-set! peg-disks 2 '())
  (send canvas refresh-now))

(define timer
  (new timer% [notify-callback move-disk]))

(reset)

(send main-frame show #t)