#lang racket

; Lexer ----------------------------------------------------------------

(define regex-ident #rx"^[A-Za-z](?>[A-Za-z]|[0-9])*")
(define regex-number #rx"^[0-9]+(?>\\.[0-9]+)?(?>e[+-]?[0-9]+)?i?")
(define regex-op #rx"^[-+*/^=]")
(define regex-ws #rx"^ *")
(define regex-fname #rx"^(sin|cos|tan|asin|acos|atan|log|ln|sqrt)")
(define regex-keyword #rx"^(let|reset|\\?)")

(define (tokenize instr)
  (let loop ([i 0])
    (let* ([str-len (string-length instr)]
           [next-pos 0]
           [start (cdar (regexp-match-positions regex-ws instr i))])
      (define (match-reg regex)
        (let ([v (regexp-match-positions regex instr start)])
          (if (equal? v #f)
              (set! next-pos #f)
              (set! next-pos (cdar v)))
          next-pos))
      (define (classify type)
        (let ([val (substring instr start next-pos)])
          (if (equal? type 'number) 
              (cons type
                    (if (equal? #\i (string-ref val (sub1 (string-length val))))
                        (string->number (string-append "0+" val))
                        (string->number val)))
              (cons type val))))
      (define (at-end)
        (or (= str-len next-pos)
            (let ([c (string-ref instr next-pos)])
              (not (or (char-numeric? c) (char-alphabetic? c))))))
      (let ([token
             (cond [(= start str-len)'()]
                   [(and (match-reg regex-keyword) (at-end))
                    (classify 'keyword)]
                   [(and (match-reg regex-fname) (at-end))
                    (classify 'fname)]
                   [(match-reg regex-ident) (classify 'ident)]
                   [(match-reg regex-number) (classify 'number)]
                   [(match-reg regex-op) (classify 'op)]
                   [(equal? #\( (string-ref instr start))
                    (set! next-pos (add1 start))
                    (cons 'lpar "(")]
                   [(equal? #\) (string-ref instr start))
                    (set! next-pos (add1 start))
                    (cons 'rpar ")")]
                   [else #f])])
        (cond [(equal? token '()) '()]
              [token (cons token (loop next-pos))]
              [else (error (format "Invalid token at ~a." start))])))))

; Parser ----------------------------------------------------------------

(define token-symbol #f)
(define token-value null)
(define token-list '())

(define (assign?) (equal? token-value "="))

(define (pow?) (equal? token-value "^"))

(define (neg?) (equal? token-value "-"))

(define (term?) (or
                 (equal? token-value "*")
                 (equal? token-value "/")))

(define (expr?) (or
                 (equal? token-value "+")
                 (equal? token-value "-")))

(define (next-symbol)
  (unless (null? token-list)
    (let ([token (car token-list)])
      (set! token-symbol (car token))
      (set! token-value (cdr token))
      (set! token-list (cdr token-list)))))

(define (accept sym)
  (if (equal? token-symbol sym)
      (begin
        (next-symbol)
        #t)
      #f))

(define (expect sym)
  (if (accept sym)
      #t
      (if (null? token-list)
          (error "Unexpected end of input.") 
          (error (format "Unexpected symbol '~a' in input." token-value)))))

(define (factor)
  (let ([val token-value])
    (cond [(accept 'number) (cons 'number val)]
          [(accept 'ident) (cons 'ident val)]
          [(accept 'lpar)
           (let ([v (expr)])
             (expect 'rpar)
             v)]
          [(accept 'fname)
           (let ([fname val])
             (expect 'lpar)
             (let ([v (expr)])
               (expect 'rpar)
               (cons 'func-call (cons fname v))))]
          [else (error (format "Invalid token: ~a." token-value))])))

(define (pow)
  (let ([e1 (factor)])
    (if (pow?)
        (begin
          (next-symbol)
          (let ([e2 (pow)])
            (cons '^ (cons e1 e2))))
        e1)))

(define (neg)
  (if (neg?)
      (begin
        (next-symbol)
        (cons 'neg (neg)))
      (pow)))

(define (term)
  (let ([e1 (neg)])
    (let loop ()
      (when (term?)
        (let ([op (if (equal? token-value "*") '* '/)])
          (next-symbol) 
          (let ([e2 (neg)])
            (set! e1 (cons op (cons e1 e2)))))
        (loop)))
    e1))

(define (expr)
  (let ([e1 (term)])
    (let loop ()
      (when (expr?)
        (let ([op (if (equal? token-value "+") '+ '-)])
          (next-symbol) 
          (let ([e2 (term)])
            (set! e1 (cons op (cons e1 e2)))))
        (loop)))
    e1))
              
(define (statement)
  (cond [(equal? token-value "let")
         (next-symbol)
         (let ([id token-value])
           (accept 'ident)
           (if (assign?)
               (begin
                 (next-symbol)
                 (cons 'assign (cons id (expr))))
               (error "Invalid let statement")))]
        [(equal? token-value "reset") (cons 'reset null)]
        [(equal? token-value "?") (cons 'help null)]
        [else (expr)]))

(define (parse instr)
  (set! token-list (tokenize (string-trim instr)))
  (next-symbol)
  (let ([val (statement)])
    (if (equal? token-list '())
        val
        (error "Syntax error in input."))))

(define (print-tree ast)
  (let loop ([level 0][node ast])
    (let ([indent (make-string (* 4 level) #\ )]
          [sym (car node)]
          [val (cdr node)])
      (printf indent)
      (define (print-op)
        (printf "Operator: ~a\n"  sym)
        (loop (add1 level) (car val))
        (loop (add1 level) (cdr val)))
      (match sym
        ['number (printf "Number: ~a\n" val)]
        ['ident  (printf "Identifier: ~a\n" val)]
        ['func-call 
         (printf "Function: ~a\n" (car val))
         (loop (add1 level) (cdr val))]
        ['+ (print-op)]
        ['- (print-op)]
        ['* (print-op)]
        ['/ (print-op)]
        ['^ (print-op)]
        ['neg 
         (printf "Neg:\n")
         (loop (add1 level) val)]
        ['assign
         (printf "Assign: ~a\n" (car val))
         (loop (add1 level) (cdr val))]
        [_ (printf "Node: ~a\n" node)]))))

(struct op (prec assoc))

(define get-prop
  (let ([op-prop
         (hash
          'assign (cons 0 'r)
          '+   (op 10 'l) 
          '-   (op 10 'l) 
          '*   (op 20 'l) 
          '/   (op 20 'l)
          'neg (op 30 'n)
          '^   (op 40 'r)
          'expt   (op 40 'r)
          'number (op 90 'n)
          'ident  (op 90 'n)
          'func-call (op 90 'n))])
    (lambda (sym)
      (hash-ref op-prop sym (lambda () (op 90 'n))))))

(define (ast->string ast)
  (let ([expr-port (open-output-string)])
    (define (push str)
      (display str expr-port))
    (let loop ([node ast])
      (let* ([sym (car node)]
             [val (cdr node)]
             [prop (get-prop sym)]
             [prec (op-prec prop)]
             [assoc (op-assoc prop)])
        (define (need-paren arg side)
          (let ([arg-prec (op-prec (get-prop (car arg)))])
            (cond [(< arg-prec prec) #t]
                  [(> arg-prec prec) #f]
                  [else (not (equal? assoc side))])))
        (define (push-op) 
          (let* ([e1 (car val)]
                 [par1 (need-paren e1 'l)]
                 [e2 (cdr val)]
                 [par2 (need-paren e2 'r)])
            (when par1 (push "("))
            (loop e1)
            (when par1 (push ")"))
            (push (format " ~a "  sym))
            (when par2 (push "("))
            (loop e2)
            (when par2 (push ")"))))
        (match sym
          ['number (push (number->string val))]
          ['ident (push val)]
          ['func-call 
           (push (car val))
           (push "(")
           (loop (cdr val))
           (push ")")]
          ['+ (push-op)]
          ['- (push-op)]
          ['* (push-op)]
          ['/ (push-op)]
          ['^ (push-op)]
          ['neg 
           (push "-")
           (let ([paren (need-paren val 'n)])
             (when paren (push "("))
             (loop val)
             (when paren (push ")")))]
          ['assign
           (push (format "let ~a = " (car val)))
           (loop (cdr val))]
          [_ (push (format "Node: ~a" sym))])))
    (get-output-string expr-port)))

(define var
  (let ([vars (make-hash)])
    (lambda args
      (match args
        [(list 'set v n) (hash-set! vars v n)]
        [(list 'get v)
         (if (hash-has-key? vars v)
             (hash-ref vars v)
             (error (format "Undefined variable: ~a" v)))]))))
    
(define (eval-ast ast)
  (let loop ([node ast])
    (let ([sym (car node)]
          [val (cdr node)])
      (define (eval-op)
        (let ([n1 (loop (car val))]
              [n2 (loop (cdr val))])
          (match sym
            ['+ (+ n1 n2)]
            ['- (- n1 n2)]
            ['* (* n1 n2)]
            ['/ (/ n1 n2)]
            ['^ (expt n1 n2)])))
      (define (eval-func fname val)
        (match fname
          ["sin" (sin (* val trig-mode))]
          ["cos" (cos (* val trig-mode))]
          ["tan" (tan (* val trig-mode))]
          ["asin" (/ (asin val) trig-mode)]
          ["acos" (/ (acos val) trig-mode)]
          ["atan" (/ (atan val) trig-mode)]
          ["ln" (log val)]
          ["log" (log val 10)]
          ["sqrt" (sqrt val)] ))
      (match sym
        ['number
         (if (and (= num-mode FLOAT) (exact? val))
             (exact->inexact val)
             val)]
        ['ident (var 'get val)]
        ['+ (eval-op)]
        ['- (eval-op)]
        ['* (eval-op)]
        ['/ (eval-op)]
        ['^ (eval-op)]
        ['neg (- (loop val))]
        ['reset (reset)]
        ['help (print-help)]
        ['assign
         (var 'set (car val)                    
              (let ([n (loop (cdr val))])
                (cond [(equal? (car val) "TrigMode")
                       (if (or (= n RADIANS) (= n DEGREES))
                           (begin
                             (set! trig-mode n)
                             (printf "TrigMode set to ~a.\n\n" (if (= n RADIANS) "Rad" "Deg")))
                           (error "Invalid TrigMode."))]
                      [(equal? (car val) "NumMode")
                       (if (or (= n FRAC) (= n FLOAT))
                           (begin
                             (set! num-mode n)
                             (printf "NumMode set to ~a.\n\n" (if (= n FRAC) "Frac" "Float")))
                           (error "Invalid NumMode."))]
                      [else n])))]
        ['func-call 
         (eval-func (car val)
                    (loop  (cdr val)))]
        [_ (error "Unknown symbol")]))))

(define RADIANS 1)
(define DEGREES (/ pi 180))
(define trig-mode RADIANS)

(define FRAC 1)
(define FLOAT 2)
(define num-mode FLOAT)

(define (reset)
  (var 'set "pi" pi)
  (var 'set "e" (exp 1))
  
  (var 'set "Rad" RADIANS)
  (var 'set "Deg" DEGREES)
  (set! trig-mode RADIANS)
  
  (var 'set "Frac" FRAC)
  (var 'set "Float" FLOAT)
  (set! num-mode FLOAT)
  
  (displayln "*** Welcome to TRAC! ***\n")
  (displayln "  Modes: Rad, Float")
  (displayln "  Enter ? for help.\n")
  )

(define (start)
  (reset)
  (let loop ()
    (display "> ")
    (with-handlers ([exn:fail? (lambda (e) (displayln (exn-message e)))])
      (let ([v (eval-ast (parse (read-line)))])
        (when (number? v) (displayln v))))
    (loop)))

#;(define (start)
    (reset)
    (let loop ()
      (display "> ")
      (with-handlers ([exn:fail? (lambda (e) (displayln (exn-message e)))])
        (let ([v (eval-ast (parse (read-line)))])
          (when (number? v)
            (if (not (real? v))
                (error "Result undefined.")
                (displayln v)))))
      (loop)))

(define (print-help)
  (let ([help (list
               (format "Current TrigMode: ~a"
                       (if (= trig-mode RADIANS) "Rad" "Deg"))              
               "To change TrigMode: to radians type:"
               "   let TrigMode = Rad"
               "To change TrigMode to degrees type:"
               "   let TrigMode = Deg"
               ""   
               (format "Current NumMode: ~a"
                       (if (= num-mode FLOAT) "Float" "Frac"))
               "To change NumMode to float type:"
               "   let NumMode: = Float"
               "To change NumMode: to fraction type:"
               "   let NumMode: = Frac"
               ""
               "To reset TRAC to defaults type:"
               "   reset")])
    (let loop([h help])
      (unless (equal? h '())
        (printf "~a\n" (car h))
        (loop (cdr h)))))
  (newline))
      
; Test code ----------------------------------------------------------

(define ops
  (vector
   (cons '+ 2) 
   (cons '- 1) ; unary minus
   (cons '- 2) ; subtraction
   (cons '* 2) 
   (cons '/ 2) 
   (cons 'expt 2) 
   (cons 'sin 1) 
   (cons 'cos 1) 
   (cons 'tan 1) 
   (cons 'asin 1) 
   (cons 'acos 1) 
   (cons 'atan 1) 
   (cons 'sqrt 1) 
   (cons 'log 1) ; natural log
   (cons 'log 2) ; base n log
   ))

(define (gen-racket)
  (let ([num-ops (vector-length ops)])
    (let loop ([d (random 1 5)])
      (if (= d 0)
          (exact->inexact (* (random) 1000000000))
          (let* ([op (vector-ref ops (random num-ops))]
                 [sym (car op)]
                 [args (cdr op)]
                 [next-d (sub1 d)])
            (if (= args 1)
                (list sym (loop next-d))
                (if (equal? sym 'log)
                    (list sym (loop next-d) 10)
                    (list sym (loop next-d) (loop next-d)))))))))

(define (racket->trac expr)
  (let ([out-port (open-output-string)])
    (define (push str)
      (display str out-port))
    (let loop ([node expr])
      (if (number? node)
          (push (number->string node))
          (let* ([sym (car node)]
                 [sym (cond [(equal? sym 'expt) '^]
                            [(equal? sym 'log)
                             (if (= (length node) 2) 'ln 'log)]
                            [(equal? sym '-)
                             (if (= (length node) 2) 'neg '-)]
                            [else sym])]
                 [prop (get-prop sym)]
                 [prec (op-prec prop)]
                 [assoc (op-assoc prop)])
            (define (need-paren arg side)
              (if (not (list? arg))
                  #f
                  (let ([arg-prec (op-prec (get-prop (car arg)))])
                    (cond [(< arg-prec prec) #t]
                          [(> arg-prec prec) #f]
                          [else (not (equal? assoc side))]))))
            (define (push-op)
              (let* ([e1 (second node)]
                     [e2 (third node)]
                     [par1 (need-paren e1 'l)]
                     [par2 (need-paren e2 'r)])
                (when par1 (push "("))
                (loop e1)
                (when par1 (push ")"))
                (push (format " ~a "  sym))
                (when par2 (push "("))
                (loop e2)
                (when par2 (push ")"))))
            (define (push-neg)
              (let* ([e (second node)]
                     [paren (need-paren e 'n)])
                (push "-")
                (when paren (push "("))
                (loop e)
                (when paren (push ")"))))
            (define (push-func)
              (push (format "~a"  sym))
              (push "(")
              (loop (second node))
              (push ")"))
            (match sym
              ['+ (push-op)]
              ['- (push-op)]
              ['* (push-op)]
              ['/ (push-op)]
              ['^ (push-op)]
              ['neg  (push-neg)]
              ['sin  (push-func)]
              ['cos  (push-func)]
              ['tan  (push-func)]
              ['asin (push-func)]
              ['acos (push-func)]
              ['atan (push-func)]
              ['ln   (push-func)]
              ['log  (push-func)]
              ['sqrt (push-func)])))
      (get-output-string out-port))))

(define (test n)
  (for ([in-range n])
    (let* ([expr (gen-racket)]
           [v1 (eval expr)]
           [v2 (eval-ast (parse (racket->trac expr)))]
           [delta (magnitude (- v1 v2))])
      (when (> delta 0)
        (displayln "Mismatch:")
        (printf "Racket: ~a\n" expr)
        (printf "TRAC: ~a\n" (racket->trac expr))
        (printf "v1: ~a, v2: ~a, delta: ~a\n\n" v1 v2 delta)))))

(start)
;(define tree (parse "(x1*cos(45) + 25 *(4e-12 / alpha))^2"))
;(print-tree tree)
;(parse "(x1*cos(45) + 25 *(4e-12 / alpha))^2")
;(tokenize "(x1*cos(45) + 25 *(4e-12 / alpha)^2")
;(tokenize "let a=4")
